% [Q_D,mu_opt] = pac_bayes_bound(gamma, delta, npriors, eta, etawrw)
%
% Computes de PAC-Bayes Bound for a linear classifier. The prior
% is a spherical Gaussian. In the classic setup of a
% Gaussian centred on the origin, only gamma and delta are
% needed to be input. A more informative prior consisting of a
% spherical Gaussian centred on the direction given by unitary wr
% at a distance eta of the origin can be specified with arguments
% eta and etawrw 
%
% Input:
%
% gamma: Normalized margin of the classifier: 
%         gamma(i) = y(i)* w'* x(i) / (|w| |x(i)|)
% delta: probability that the bound does not hold
% npriors: number of priors, in case the function is called from a prior
%          pac-bayes bound
% eta: length of the weight vector of the prior distribution
% etawrw: dot product between the prior and a unitary posterior:
%         eta*wr'*w / (|w|) where eta*wr is a prior defined as a
%         spherical Gaussian centred along the direction given by 
%         wr at a distance eta from the origin and w/|w| is the 
%         unitary posterior.
%         
% Output:
%
% Q_D: Expected True Error of the stochastic classifier
% mu_opt: value of mu that yields the minimum Q_D 
% 
% EPH, G2PI/DTSC/UC3M, June 20th, 2006.
%

function [Q_D,mu_opt] = pac_bayes_bound(gamma,delta, npriors, eta, etawrw)
  
  
  if nargin < 3
    npriors = 1;
    eta=0;
    etawrw=0;
  end
  
  mr = length(gamma);
  ln = log((mr+1)/delta) + log(npriors);
  Q_D = 100; % initial value for Q_D. It will be <1...
  mu = [.0001 1 50 100]; %initial values for mu
  mu_opt=1;                       
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % 1.- Compute stochastic error Q_S
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  mu_gamma = kron(mu,gamma); 
  F = 1-normcdf(mu_gamma);
  Qs = mean(F); % training stochastic error
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % 2.- Compute the KL divergence between Prior wr and Posterior
  %     w. This corresponds to ||eta wr- mu w||^2
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  KLPQ = mu.^2 - 2*mu*etawrw + eta*eta;
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % 3.- Right hand side of the bound, i.e., KL divergence between true
  % error and empirical error
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  kl = (KLPQ/2 + ln)/(mr); % kl divergence between Qd and Qs (true
                           % and stochastic errors)
  qd = ones(1,length(kl));
  ccual = find(Qs<.999999);
  qd(ccual) = inv_kl(Qs(ccual),kl(ccual)); % quick and dirty
                                           % inversion of the KL divergence
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Loop to obtain the optimal value for mu
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   
  while (((mu(end)-mu(1))/mu(end)) > .05 ) 
    % stop condition, make the interval where mu lives narrow enough 
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 4.- Linear search to invert the kl divergence and find best mu
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    [pas cual] = min(qd);
    if pas < Q_D
      Q_D = pas;
      mu_opt = mu(cual);
    end
    
    % refine the four values for mu
    switch cual
     case 1,
      mu=[mu(1)/4 mu(1)/2 mu(1:2)];
      qd=[100 100 qd(1:2)];
      nuevos=[1 2];
     case 2,
      if qd(1)>qd(3)
        mu=[mu(1) (mu(1)+mu(2))/2 mu(2) mu(3)];
        qd=[qd(1) 100 qd(2:3)];
        nuevos=2;
      else
        mu=[mu(1)  mu(2) (mu(3)+mu(2))/2 mu(3)];
        qd=[qd(1:2) 100 qd(3)];
        nuevos=3;
      end
     case 3,
      if qd(2)>qd(4)
        mu=[mu(2) (mu(2)+mu(3))/2 mu(3) mu(4)];
        qd=[qd(2) 100 qd(3:4)];
        nuevos=2;
      else
        mu=[mu(2)  mu(3) (mu(3)+mu(4))/2 mu(4)];
        qd=[qd(2:3) 100 qd(4)];
        nuevos=3;
      end
     case 4,
      mu=[mu(3) mu(4) mu(4)*2 mu(4)*4];
      qd=[qd(3:4) 100 100];
      nuevos=[3,4];
    end
    mu_gamma = kron(mu(nuevos),gamma); 
    F = 1-normcdf(mu_gamma);
    Qs = mean(F); % training stochastic error
    KLPQ = mu(nuevos).^2 - 2*mu(nuevos)*etawrw + eta*eta;

    kl = (KLPQ/2 + ln)/(mr); % kl divergence between errors
    qdd = ones(1,length(kl));
    ccual = find(Qs<.999999);
    qdd(ccual) = inv_kl(Qs(ccual),kl(ccual));
    qd(nuevos) = qdd;
    
  end %while  
  
  
  
  
