% [Q_D,mu_opt] = taupsvm_bound(gamma,delta,tau,eta0,wrw)
%
% Computes de Prior PAC-Bayes Bound for a SVM. The prior
% distribution on classifiers is a Gaussian centred along wr, at a
% distance eta0 from the origin and with its covariance matrix
% stretched a quantity of tau along the direction given by wr 
%
% Input:
%
% gamma: Normalized margin of the classifier: 
%         gamma(i) = y(i)* w'* x(i) / (|w| |x(i)|)
% delta: probability that the bound does not hold
% tau: varianze of the prior in the direction w_r
% eta0: distance from the centre of the Gaussian prior to the origin. This has to be fixed before knowing the length
%       of the weight vector of the prior distribution determined 
%       after the optimisation
% wrw: w_r'*w / (|w_r| |w|) where w_r/|w_r| is the normalized prior
%          weight of classifiers, w/|w| is the normalized posterior
% Output:
%
% Q_D: True error of the stochastic classifier
% mu_opt: value of mu that yields the minimum Q_D  the search
% 
% EPH, G2PI/DTSC/UC3M, June 29th, 2006.
%

function [Q_D,mu_opt] = taupsvm_bound(gamma, delta, tau, eta0, wrw)  
  
  if nargin < 4
    wrw=0;
  end
  
  tau2 = tau*tau;
  
  mr = length(gamma); % m-r in the paper
  ln = log((mr+1)/delta);
  Q_D = 100; % initial value for Q_D
  mu = [.0001 1 50 100]; %initial values for mu
  mu_opt=1;                       
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % 1.- Compute stochastic error Q_S
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  mu_gamma = kron(mu,gamma); 
  F = 1-normcdf(mu_gamma);
  Qs = mean(F); % training stochastic error
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % 2.- Compute the KL divergence between Prior wr and Posterior
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  KLPQ = (log(tau2) + (1/tau2 - 1) + (mu*wrw - eta0).^2/tau2 + mu.^2 * (1-wrw^2));
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % 3.- Right hand side of the bound, i.e., KL divergence between true
  % error and empirical error
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  kl = (KLPQ/2 + ln)/(mr); % kl divergence between errors
  qd = ones(1,length(kl));
  ccual = find(Qs<.999999);
  qd(ccual) = inv_kl(Qs(ccual),kl(ccual));
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  % Loop to obtain the optimal value for mu
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   
  while (((mu(end)-mu(1))/mu(end)) > .05 ) 
    % stop condition, make the interval where mu lives narrow enough 
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % 4.- Linear search to invert the kl divergence and find best mu
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    [pas cual] = min(qd);
    if pas < Q_D
      Q_D = pas;
      mu_opt = mu(cual);
    end
    
    % refine the four values for mu
    switch cual
     case 1,
      mu=[mu(1)/4 mu(1)/2 mu(1:2)];
      qd=[100 100 qd(1:2)];
      nuevos=[1 2];
     case 2,
      if qd(1)>qd(3)
        mu=[mu(1) (mu(1)+mu(2))/2 mu(2) mu(3)];
        qd=[qd(1) 100 qd(2:3)];
        nuevos=2;
      else
        mu=[mu(1)  mu(2) (mu(3)+mu(2))/2 mu(3)];
        qd=[qd(1:2) 100 qd(3)];
        nuevos=3;
      end
     case 3,
      if qd(2)>qd(4)
        mu=[mu(2) (mu(2)+mu(3))/2 mu(3) mu(4)];
        qd=[qd(2) 100 qd(3:4)];
        nuevos=2;
      else
        mu=[mu(2)  mu(3) (mu(3)+mu(4))/2 mu(4)];
        qd=[qd(2:3) 100 qd(4)];
        nuevos=3;
      end
     case 4,
      mu=[mu(3) mu(4) mu(4)*2 mu(4)*4];
      qd=[qd(3:4) 100 100];
      nuevos=[3,4];
    end
    mu_gamma = kron(mu(nuevos),gamma); 
    F = 1-normcdf(mu_gamma);
    Qs = mean(F); % training stochastic error
    KLPQ = (log(tau2) + (1/tau2 - 1) + (mu(nuevos)*wrw - eta0).^2/tau2 + mu(nuevos).^2 * (1-wrw^2));
    kl = (KLPQ/2 + ln)/(mr); % kl divergence between errors
    qdd = ones(1,length(kl));
    ccual = find(Qs<.999999);
    qdd(ccual) = inv_kl(Qs(ccual),kl(ccual));
    qd(nuevos) = qdd;
    
  end %while  
  
  
  
  
